#!/usr/bin/perl -w
use strict;
use Getopt::Long;
use File::Basename;
use List::Util qw[min max];

###############################################################
##################### DOCUMENTATION ###########################

## Author: Yun Li
## Date: 2010-11-19
## Goal: Split pedigree files into batches with smaller number of markers
my $location = $0;
my $title = "splitPed by marker Version 0.0.1";

print "$title\n";
print "\t\@: $location\n";
print "\tby: Yun Li(ylwtx\@med.unc.edu)\n\n";

## Input Parameters
## splitPed (default, no)
## nWindows (# windows to split into), still need overlapSize
## windowSize (#SNPs)
## overlapSize (#SNPs)
## windowLength (bp), works only when map input file exisits
## overlapLength (bp), works only when map input file exists
## extimateWindowOnly (default, no)

## Input Files
## Required
##    (*) .ped
##    (*) .dat
## Optional
##    (*) .map 

## Output Files

## NOTES 
## (*) S2 markers will not be filtered out
## (*) assume markers are in the right order: does not check map file when provided
## (*) deals with heading A or T rows in .dat: will be discarded in output if separate ped files are generated
## (*) count any fields after as marker columns. So A and T mixed inside will NOT be handled properly.

################# END OF DOCUMENTATION ########################
###############################################################
my $starttime = time();

my $MIN_NUM_OPTS = 5;

if($#ARGV<($MIN_NUM_OPTS*2 - 1))
{
	&usage();
	die "Failed to parse parameters\n\n";
}

my %opts = ();

	# Default Optional Options
$opts{splitPed} = 0;
$opts{extimateWindowOnly} = 0;
  # other parameters
  $opts{ped}=0;
  $opts{dat}=0;
  $opts{nWindows}=0;
  $opts{windowSize}=0;
  $opts{windowLength}=0;
  $opts{overlapSize}=0;
  $opts{overlapLength}=0;
  $opts{map}=0;

Getopt::Long::GetOptions(\%opts,qw(
	splitPed=s
	nWindows=i
	windowSize=i
	overlapSize=i
	windowLength=i
	overlapLength=i
	extimateWindowOnly=s
	ped=s
	dat=s
	map=s
	o=s
)) || die "Failed to parse options\n\n";

&printPar();

my $FIXED_HEADING = 5;
my $nmarkers; # total number of records in .dat file (including S2, A etc);
my $totalLength;
my $headingNonMarkerColumns = 0;
&main();

my $endtime = time();
my $lapsetime = $endtime - $starttime;
print "\nAnalysis took $lapsetime seconds\n\n";

# begin sub

sub usage
{
	print "\n";
	print "Usage:\n\t";
	print "-splitPed\t If .ped file is splitted? (Required, default=0) \n\t";
	print "-nWindows\t # windows to split into (Co-Required Set 1) \n\t";
	print "-windowSize\t window size in terms of # of markers per window (Co-Required Set 1) \n\t";
	print "-windowLength\t window size in terms of # of base pairs (Co-Required Set 1, and can only be used when map file specified) \n\t";
	print "-overlapSize\t # markers overlapping between two consecutive windows (Co-Required Set 2) \n\t";
	print "-overlapLength\t # base pairs overlapping between two consecutive windows (Co-Required Set 2) \n\t";
	print "-extimateWindowOnly\t Only estimate # windows, window size/length, and overlapping size/length (Optional, default=0) \n\t";
	print "-ped\t Input pedigree file (Required) \n\t";
	print "-dat\t Input dat (marker info) file (Required) \n\t";
	print "-map\t Input map file (Optional) \n\t";
	print "-o\t Output Prefix (Required) \n\t";
	print "\n";
}

sub printPar
{
	print "\n";
	print "Parameters in Effect:\n";
	print "\t splitPed \n\t\t-splitPed '$opts{splitPed}'\n";
	print "\t nWindows \n\t\t-nWindows '$opts{nWindows}'\n";
	print "\t windowSize \n\t\t-windowSize '$opts{windowSize}'\n";
	print "\t windowLength \n\t\t-windowLength '$opts{windowLength}'\n";
	print "\t overlapSize \n\t\t-overlapSize '$opts{overlapSize}'\n";
	print "\t overlapLength \n\t\t-overlapLength '$opts{overlapLength}'\n";
	print "\t extimateWindowOnly \n\t\t-extimateWindowOnly '$opts{extimateWindowOnly}'\n";
	print "\t ped file \n\t\t-ped '$opts{ped}'\n";
	print "\t dat file \n\t\t-dat '$opts{dat}'\n";
	print "\t map file \n\t\t-map '$opts{map}'\n";
	print "\t Output Prefix \n\t\t-o '$opts{o}'\n";
	print "\n";
}

sub rm
{
	if(-e $_[0])
	{
		print "WARNING: $_[0] existed and removed!!\n\n";
		system "/bin/rm -rf $_[0]";
	}
}

sub bakrm
{
	if(-e $_[0])
	{
		my $tt = $_[0].".old";
		print "WARNING: $_[0] existed and moved to $tt !!\n\n";
		system "mv $_[0] $tt";
	}
}

sub swap
{
	my $tmp = $_[0];
	$_[0] = $_[1];
	$_[1] = $tmp;
}

#==================================================================
# Subroutine:
#   Open($filename, $handle)
#
# Usage:
#   Open('myfile.txt', \*IN);
#   Open($pedfile, \*PED);
#
#   Then refernce the handles <IN> and <PED> as usual.
#
# Description:
#   Open the file 'filename'. The file can be a normal file
#   or one compressed with gzip or bzip2.
#   Function does not return if the open fails.
#
# Arguments:
#   $filename - path to file to open
#   $handle - GLOB name of handle to use. Include '\*'
#
#==================================================================
sub Open {
    my ($filename, $handle) = @_;

    # Handle gzipped data
    if ($filename =~ /\.gz$/) {
        my $cmd = "gunzip -c $filename";
        open($handle, "$cmd |") ||
            die "Unable to invoke command '$cmd': $!\n";
        if (eof($handle)) { die "Cmd failed: '$cmd'\n"; }
        return 1;
    }
    # Handle bzip2 data
    if ($filename =~ /\.bz2$/) {
        my $cmd = "bunzip2 -c $filename";
        open($handle, "$cmd |") ||
            die "Unable to invoke command '$cmd': $!\n";
        if (eof($handle)) { die "Cmd failed: '$cmd'\n"; }
        return 2;
    }
    # Handle normal file
    open($handle, $filename) ||
        die "Unable to open file '$filename': $!\n";
    return 3;
}


sub main
{
  ###################
  ## sanity checks ##
  ###################
  if ($opts{ped} eq 0 || $opts{dat} eq 0 || $opts{o} eq 0)
  {
    print "ERROR: ped, dat, output prefix are all required.\n\n";
    exit;
  }
  
  if ($opts{nWindows} == 0 && $opts{windowSize} == 0 && $opts{windowLength} == 0)
  {
    print "ERROR: window size must be specified, either by nWindows, or by windowSize, or by windowLength.\n\n";
    exit;
  }
  
  if ($opts{overlapSize} == 0 && $opts{overlapLength} == 0)
  {
    print "ERROR: overlapping window size must be specified, either by overlapSize, or by overlapLength.\n\n";
    exit;
  }
  
  if ($opts{map} eq 0)
  {
    if ($opts{overlapLength} != 0 || $opts{windowLength} != 0)
    {
      print "ERROR: neither overlapLength nor windowLength is allowed because map file not specified.\n\n";
      exit;
    }
  }
  
  #####################################
  ## find heading non-marker columns ##
  #####################################
  open(IN, $opts{dat}) || die "Cannot open $opts{dat}\n\n";
  while (my $line = <IN>)
  {
    chomp($line);
    my $flag = (split(/\s+/,$line))[0];
    if ($flag eq "A" || $flag eq "T")
    {
      $headingNonMarkerColumns ++
    }
    else
    {
      last;
    }
  }
  print "There are $headingNonMarkerColumns heading columns not for markers\n";
  
  ##########################
  ## estimate window size ##
  ##########################
  &estimateWindow();
  
  if ($opts{extimateWindowOnly} ne 0)
  {
    exit;
  }
  
  #############################
  ## split the pedigree file ##
  #############################
  &split();
  
}

sub estimateWindow
{
  #########################
  ## part 1: core window ##
  #########################
  my $coreMarkers;
  my $nWindows;
  my $endPos;
  my $startPos;
  
  $nmarkers = `wc -l < $opts{dat}`;
  chomp($nmarkers);
  $nmarkers -= $headingNonMarkerColumns;
  ######################################
  ## scenario 1.1: nWindows specified ##
  ######################################
  if ($opts{nWindows} != 0)
  {
    if ($opts{windowSize} != 0 || $opts{windowLength} != 0)
    {
      print "ERROR: Specify one and only one of the following three: nWindows, windowSize, windowLength\n\n";
      exit;
    }
    $nWindows = $opts{nWindows};
    $coreMarkers = int($nmarkers / $opts{nWindows});
  }
  
  ########################################
  ## scenario 1.2: windowSize specified ##
  ########################################
  if ($opts{windowSize} != 0)
  {
    if ($opts{nWindows} != 0 || $opts{windowLength} != 0)
    {
      print "ERROR: Specify one and only one of the following three: nWindows, windowSize, windowLength\n\n";
      exit;
    }
    $coreMarkers = $opts{windowSize};
    $nWindows = $nmarkers / $opts{windowSize};
    if ( ($nWindows - int($nWindows)) > 0.5) {$nWindows = int($nWindows) + 1;}
    else {$nWindows = int($nWindows); }
  }
  
  ##########################################
  ## scenario 1.3: windowLength specified ##
  ##########################################
  if ($opts{map} ne 0)
  {
    $endPos = `tail -n 1 $opts{map} | awk '{print \$3}'`;
    chomp($endPos);
    $startPos = `head -1 $opts{map} | awk '{print \$3}'`;
    chomp($startPos);
    $totalLength = $endPos + 1 - $startPos;
  }
  if ($opts{windowLength} != 0)
  {
    if ($opts{windowSize} != 0 || $opts{nWindows} != 0)
    {
      print "ERROR: Specify one and only one of the following three: nWindows, windowSize, windowLength\n\n";
      exit;
    }
    $nWindows = $totalLength / $opts{windowLength};
    if ( ($nWindows - int($nWindows)) > 0.5) {$nWindows = int($nWindows) + 1;}
    else {$nWindows = int($nWindows); }
  }
  
  #############################
  ## part 2: flanking window ##
  #############################
  my $flankingMarkers;
  #########################################
  ## scenario 2.1: overlapSize specified ##
  #########################################
  if ($opts{overlapSize} != 0)
  {
    if ($opts{overlapLength} != 0)
    {
      print "ERROR: Specify one and only one of the following two: overlapSize, overlapLength\n\n";
      exit;
    }
    $flankingMarkers = $opts{overlapSize};
  }
  
  ###########################################
  ## scenario 2.2: overlapLength specified ##
  ###########################################
  if ($opts{overlapLength} != 0)
  {
    if ($opts{overlapSize} != 0)
    {
      print "ERROR: Specify one and only one of the following two: overlapSize, overlapLength\n\n";
      exit;
    }
  }

  ##################################
  ## part 3: print the statistics ##
  ##################################
  if ($opts{overlapLength} == 0 && $opts{windowLength} ==0)
  {
    print "\nWill split total $nmarkers SNPs into $nWindows windows, each with [$coreMarkers + 2 * $flankingMarkers ] markers in window\n\n";
  }
  else
  {
    print "\nWill split total $totalLength base pairs into the following $nWindows windows:\n";
    print "window [windowStart (coreStart coreEnd) windowEnd] #leftFlankingSNPs #coreSNPs #rightFlankingSNPs #totalSNPs\n";
    if ($opts{overlapLength} != 0 && $opts{windowLength} !=0)
    {
      for (my $i = 0; $i < $nWindows; $i ++)
      {
        my $windowIndex = $i + 1;
        my $coreStart = $opts{windowLength} * $i + $startPos;
        my $coreEnd = $coreStart + $opts{windowLength};
        my $coreSNPs = `awk -v s=$coreStart -v e=$coreEnd '\$3 >= s && \$3 <= e' $opts{map} | wc -l`;
        chomp($coreSNPs);
        
        # the following would work for the first and last region;
        my $winStart = $coreStart - $opts{overlapLength};
        my $winEnd = $coreEnd + $opts{overlapLength};
        if ($windowIndex == $nWindows) {$winEnd = $endPos;}
        my $winSNPs = `awk -v s=$winStart -v e=$winEnd '\$3 >= s && \$3 <= e' $opts{map} | wc -l`;
        chomp($winSNPs);
        
        # check if left overlapping region contains any markers
        my $leftSNPs = `awk -v s=$winStart -v e=$coreStart '\$3 >= s && \$3 < e' $opts{map} | wc -l`;
        chomp($leftSNPs);
        
        # check if right overlapping region contains any markers
        my $rightSNPs = `awk -v s=$coreEnd -v e=$winEnd '\$3 > s && \$3 <= e' $opts{map} | wc -l`;
        chomp($rightSNPs);
        
        print "$windowIndex [$winStart ($coreStart $coreEnd) $winEnd] $leftSNPs $coreSNPs $rightSNPs $winSNPs\n";
        ## sanity checks 
        if ($leftSNPs == 0 && $windowIndex != 1)
        {
          print "ERROR: NO markers in the left flanking region for window $windowIndex\n";
          exit;
        }
        if ($rightSNPs == 0 && $windowIndex != $nWindows)
        {
          print "ERROR: NO markers in the right flanking region for window $windowIndex\n";
          exit;
        }
      }
    }
    else
    {
      print "ERROR: only one of overlapLength and windowLength is specified. Specify either both or none\n\n";
      exit;
    }
  }
}

sub split
{
  #########################
  ## part 1: core window ##
  #########################
  my $coreMarkers;
  my $nWindows;
  my $endPos;
  my $startPos;
  
  $nmarkers = `wc -l < $opts{dat}`;
  chomp($nmarkers);
  $nmarkers -= $headingNonMarkerColumns;
  ######################################
  ## scenario 1.1: nWindows specified ##
  ######################################
  if ($opts{nWindows} != 0)
  {
    if ($opts{windowSize} != 0 || $opts{windowLength} != 0)
    {
      print "ERROR: Specify one and only one of the following three: nWindows, windowSize, windowLength\n\n";
      exit;
    }
    $nWindows = $opts{nWindows};
    $coreMarkers = int($nmarkers / $opts{nWindows});
  }
  
  ########################################
  ## scenario 1.2: windowSize specified ##
  ########################################
  if ($opts{windowSize} != 0)
  {
    if ($opts{nWindows} != 0 || $opts{windowLength} != 0)
    {
      print "ERROR: Specify one and only one of the following three: nWindows, windowSize, windowLength\n\n";
      exit;
    }
    $coreMarkers = $opts{windowSize};
    $nWindows = $nmarkers / $opts{windowSize};
    if ( ($nWindows - int($nWindows)) > 0.5) {$nWindows = int($nWindows) + 1;}
    else {$nWindows = int($nWindows); }
  }
  
  ##########################################
  ## scenario 1.3: windowLength specified ##
  ##########################################
  if ($opts{map} ne 0)
  {
    $endPos = `tail -n 1 $opts{map} | awk '{print \$3}'`;
    chomp($endPos);
    $startPos = `head -1 $opts{map} | awk '{print \$3}'`;
    chomp($startPos);
    $totalLength = $endPos + 1 - $startPos;
  }
  if ($opts{windowLength} != 0)
  {
    if ($opts{windowSize} != 0 || $opts{nWindows} != 0)
    {
      print "ERROR: Specify one and only one of the following three: nWindows, windowSize, windowLength\n\n";
      exit;
    }
    $nWindows = $totalLength / $opts{windowLength};
    if ( ($nWindows - int($nWindows)) > 0.5) {$nWindows = int($nWindows) + 1;}
    else {$nWindows = int($nWindows); }
  }
  
  #############################
  ## part 2: flanking window ##
  #############################
  my $flankingMarkers;
  #########################################
  ## scenario 2.1: overlapSize specified ##
  #########################################
  if ($opts{overlapSize} != 0)
  {
    if ($opts{overlapLength} != 0)
    {
      print "ERROR: Specify one and only one of the following two: overlapSize, overlapLength\n\n";
      exit;
    }
    $flankingMarkers = $opts{overlapSize};
  }
  
  ###########################################
  ## scenario 2.2: overlapLength specified ##
  ###########################################
  if ($opts{overlapLength} != 0)
  {
    if ($opts{overlapSize} != 0)
    {
      print "ERROR: Specify one and only one of the following two: overlapSize, overlapLength\n\n";
      exit;
    }
  }

  ###################
  ## part 3: split ##
  ###################
  ## default: share the same ped file, but mask different portion of .dat file ##
  for (my $i = 0; $i < $nWindows; $i ++)
  {
    my $index = $i + 1;
    
    #############################
    ## generate output file(s) ##
    #############################
      ## create output handlers ##
      my $odf = $opts{o}.".".$index.".dat";
      #&rm($odf);
      #open(ODF, ">>$odf") || die "cannot create file $odf\n\n";
      my $opf = $opts{o}.".".$index.".ped";
      if ($opts{splitPed} ne 0)
      {
        &rm($opf);
        open(OPF, ">>$opf") || die "cannot create file $opf\n\n";
      }
      
      ## determine marker index of starting and ending SNPs in window ##
      my $startIndex;
      my $endIndex;
    
      if ($opts{overlapLength} == 0 && $opts{windowLength} ==0)
      {
        $startIndex = $i * $coreMarkers - $flankingMarkers;
        if ($i == 0) {$startIndex += $flankingMarkers;}
        $endIndex = $index * $coreMarkers + $flankingMarkers;
        if ($index == $nWindows) {$endIndex = $nmarkers - 1;}
      }
      else
      {
        if ($opts{overlapLength} != 0 && $opts{windowLength} !=0)
        {
          my $coreStart = $opts{windowLength} * $i + $startPos;
          my $coreEnd = $coreStart + $opts{windowLength};
          
          # the following would work for the first and last region;
          my $winStart = $coreStart - $opts{overlapLength};
          my $winEnd = $coreEnd + $opts{overlapLength};
          if ($index == $nWindows) {$winEnd = $endPos;}
          
          $startIndex = `cat -n $opts{map} | awk -v s=$winStart -v e=$winEnd '\$4 >= s && \$4 <= e' | awk '{print \$1}' | head -1` - 1;
          $endIndex = `cat -n $opts{map} | awk -v s=$winStart -v e=$winEnd '\$4 >= s && \$4 <= e' | awk '{print \$1}' | tail -n 1` - 1;
          
          #print "$startIndex $endIndex\n";
        }
        else
        {
          print "ERROR: only one of overlapLength and windowLength is specified. Specify either both or none\n\n";
          exit;
        }
      }
      
      ## generate output ##
      print "Generating output for window $index ...\n";
      
        ## generate dat file ##
        my $startLine = $startIndex + 1 + $headingNonMarkerColumns;
        my $endLine = $endIndex + 1 + $headingNonMarkerColumns;
        my $nLines = $endIndex - $startIndex + 1;
        if ($opts{splitPed} ne 0)
        {
          system "head -$endLine $opts{dat} | tail -n $nLines > $odf";
        } # generate a reduced dat for each window
        else
        {
          my $lastBeforeStart = $startLine - 1;
          system "head -$lastBeforeStart $opts{dat} | sed -e 's/^M/S2/g' > $odf";
          system "head -$endLine $opts{dat} | tail -n $nLines >> $odf";
          system "sed 1,${endLine}d $opts{dat} | sed -e 's/^M/S2/g' >> $odf";
        } # generate a masked dat for each window (same length as the original)
        
        ## generate ped file ##
        if ($opts{splitPed} ne 0)
        {
          Open($opts{ped},\*IN) || die "cannot open file $opts{ped}\n\n";
          my $linenum = 0;
          while(my $line = <IN>)
          {
            $linenum ++;
            chomp($line);
            my @linearr = (split(/\s+/,$line));
            my $obsfields = scalar(@linearr);
            my $expfields = $nmarkers + $headingNonMarkerColumns + $FIXED_HEADING;
            if ($obsfields != $expfields)
            {
              print "ERROR: line $linenum of opts{ped} contains $obsfields fields, while expecting $expfields based on $opts{dat}\n\n";
              exit;
            }
            
            # print out the first five columns
            for (my $k = 0; $k < $FIXED_HEADING; $k ++)
            {
              print OPF "$linearr[$k] ";
            }
            
            # print out the genotype columns
            my $pedStartIndex = $FIXED_HEADING + $headingNonMarkerColumns + $startIndex;
            my $pedEndIndex = $FIXED_HEADING + $headingNonMarkerColumns + $endIndex;
            for (my $k = $pedStartIndex; $k <= $pedEndIndex; $k ++)
            {
              print OPF "$linearr[$k] ";
            }
            print OPF "\n";
          }
          close(IN);
        } ## end of generating ped file
   
      close(OPF);
      #close(ODF);
    
    #############################
    ## generate output file(s) ##
    ##            END          ##
    #############################
    
  } # end of window i 
}  
  
  

